#!/usr/bin/python
# -*- coding: utf-8 -*- 

################################################################################
#
# Servidor Tablon - Execution manager
# Asignatura Computación paralela
# Javier Fresno, 2015
#
################################################################################


#
# Logging
#
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('tablon')

# Includes
import time
import os
import threading
from Queue import Queue

# Import the tablondb
import tablondb

#
# Instance to tablon module
#
import sys
tablon = sys.modules['tablon'] 


#
# Execution manager, it handles the execution requests and uses the queues
# to execute them
#
class ExecutionManager:

	# File to store the last request id
	LASTREQUEST_FILE = 'serverdata/lastrequest.dat'

	#
	# Run method
	#	
	def run(self):

		# Start manager
		self.running = True
		self.keep_running = True
		logger.info("Starting Execution Manager")

		# Get last id
		self.requestid = self.load_last_request()

		# Lock to add new requests
		self.lock = threading.Lock()
		# List with the input request (it needs the lock)
		self.inrequests = [] 

		# List with the request managed by the manager
		self.requests = []

		#
		# Create and start the server
		#
		self.thread = threading.Thread(target=self.run_thread)
		self.thread.start()


	#
	# Load the last request id from the file
	#
	def load_last_request(self):

		rfile = 0
		rdb = 0

		if os.path.isfile(self.LASTREQUEST_FILE):
			with open(self.LASTREQUEST_FILE, 'r') as f:
				rfile = int(f.read())

		rdb = tablondb.get_last_request()
		return max(rfile,rdb)

	#
	# Save the last request id to the file
	# 
	def save_last_request(self):
		with open(self.LASTREQUEST_FILE,'w') as f:
			f.write('%d' % self.requestid)

	#
	# Init the stop process
	#	
	def start_stop(self):
		self.keep_running = False
	
	#
	# Check if the manager is running
	#
	def is_running(self):
		return self.running

	#
	# Run method
	#		
	def run_thread(self):

		# While is running, add new request and
		# manage the requests in the list 
		while(self.keep_running):

			self.add_new_requests()
			self.manage_requests()

			# Not bussy wait
			time.sleep(0.00001)

			# Sleep a bit if we do not have task to do
			if(len(self.requests) == 0):
				time.sleep(1)

		# Cancel the requests that are still active
		self.cancel_requests()
			
		# Now it is closed
		self.running = False
		logger.info("Execution Manager frontend closed")


	# Add a request
	# This method is called by several threads
	def add(self,request):

		# Get the lock
		with self.lock:
			self.requestid += 1
			rid = self.requestid
			request.rid = rid
			self.save_last_request()
			self.inrequests.append(request)
			return rid

	#
	# move the request from the shared list to the private one
	#	
	def add_new_requests(self):

		# Use the lock because inrequests is shared
		with self.lock:
			for request in self.inrequests:
				self.requests.append(request)
			self.inrequests = [] 


	#
	# Manage the requests
	#	
	def manage_requests(self):

		# Iterate over the request and update them in the queue
		for request in self.requests:
			queue = tablon.queues[request.queue]
			queue.update(request)

		# Remove finished items
		requests = [x for x in self.requests if not x.finished()]
		self.requests = requests


	#
	# Cancel the requests
	#	
	def cancel_requests(self):

		# Iterate over the request and update them in the queue
		for request in self.requests:
			queue = tablon.queues[request.queue]
			queue.cancel(request)












