#!/usr/bin/python
# -*- coding: utf-8 -*- 

################################################################################
#
# Servidor Tablon - Execution queue
# Asignatura Computación paralela
# Javier Fresno, 2015
#
################################################################################

#
# Logging
#
import logging
logger = logging.getLogger('tablon')

#
# Instance to tablon
#
import sys
tablon = sys.modules['tablon'] 

#
# Import the tablon DB
#
import tablondb

# The Status of the requests
from executionrequest import Status
from executionthread import ExecutionThread
from leaderboard import LeaderBoard

# Imports
import os
import shutil





#
# Basic queue class with the common methods
#
class BasicQueue(object):

	# Global lock for exclusive queues
	is_other_running = False

	# Paths
	SRCPATH = 'usercodes/src'
	BINPATH = 'usercodes/bin'

	#
	# Constructor
	#
	def __init__(self,name,**kwargs):

		logger.info("Creating queue '%s'" % name)

		# Load the parameters
		self.name = name
		self.nprocs = kwargs.get('nprocs',1)
		self.threads = kwargs.get('threads',1)
		self.nprocs_x_threads = kwargs.get('nprocs_x_threads',self.nprocs*self.threads)
		# ARTURO: GPUs
		self.ngpus = kwargs.get('ngpus',0)
		self.extensions = kwargs.get('extensions',['.c'])
		self.timewall = kwargs.get('timewall',60)
		self.description = kwargs.get('description','')
		self.makefile = kwargs.get('makefile')
		self.extrafiles = kwargs.get('extrafiles',[])
		self.exclusive = kwargs.get('exclusive',True)

		# dict for the leaderboards
		self.leaderboards = dict()

		# Create the folders
		if not os.path.exists(self.SRCPATH):
			os.makedirs(self.SRCPATH)
		if not os.path.exists(self.BINPATH):
			os.makedirs(self.BINPATH)

		# Copy the makefile
		# The makefile
		if not self.makefile:
			raise Exception("I need a makefile")	
		shutil.copy('makefiles/%s' % self.makefile, self.SRCPATH)

		# Copy the other files
		# The list can be a string of a tuple if we need to change the file name
		for f in self.extrafiles:
			if isinstance(f, str):
				shutil.copy('makefiles/%s' % f, self.SRCPATH)
			else:
				shutil.copy('makefiles/%s' % f[0], self.SRCPATH + '/%s' % f[1])
				

		# Add the queue to tablon
		tablon.queues[self.name] = self
		tablon.onlyqueues.append(self)



	#
	# Make a leaderboard in the queue
	#
	def addLeaderboard(self,name,description='',**kwargs):

		logger.info("Creating leaderboard '%s' in queue '%s'" % (name,self.name))

		lb = LeaderBoard(name,description,**kwargs)
		self.leaderboards[name] = lb
		tablon.queues[name] = self
		return lb

	# Return a leaderboard
	def getLeaderboard(self,name):
		return self.leaderboards[name]

	#
	# Update the the main method that call the other tasks depending
	# on the Status of the requests.
	#
	def update(self,request):

		if request.status == Status.QUEUED:
			self.check(request)
		elif request.status == Status.COMPILING:
			self.compile(request)
		elif request.status == Status.DEPLOYING:
			self.deploy(request)
		elif request.status == Status.WAITING:
			self.execute(request)
		elif request.status == Status.RUNNING:
			self.execute_end(request)
	
	# Checks the request and adds to the queue or marks as error
	def check(self,request):
		pass

	# Compile the program
	def compile(self,request):
		pass

	# Deploys the binary file
	def deploy(self,request):
		request.status = Status.WAITING

	# Execute the code
	def execute(self,request):
		pass

	# Checks the finalization of the excution
	def execute_end(self,request):
		pass

	def cancel(self,request):
		pass

	# Check the number of procs and threads
	# used by the child classes
	def check_nprocs_threads(self,request):

		# Check number of processes and threads
		if request.nprocs > self.nprocs or request.nprocs < 1:
			return (False, "Invalid number of processors")
		if request.threads > self.threads or request.threads < 1:
			return (False, "Invalid number of threads")
		if request.nprocs*request.threads > self.nprocs_x_threads:
			return (False, "Invalid number of processors and threads")
		# ARTURO: GPUs
		#if request.ngpus > self.ngpus:
		#	return (False, "Invalid number of GPUs")

		return (True,None)



	# Check the extension
	# used by the child classes
	def check_extension(self,request):

		ext = request.extension()
		if not ext in self.extensions:
			return (False, "Invalid extension %s" % ext)
		return (True,None)


	# Check the leaderboard of the requets (if any)
	def check_leaderboard(self,request):


		# If the name of the queue is the same in the request
		# means that we are not in a leaderboard execution
		if self.name != request.queue:

			# Set leaderboard to true
			request.setLeaderboard()

			# Get the leaderboard to check if it is open
			lb = self.getLeaderboard(request.queue)
			if not lb.isopen:
				return (False,"Leaderboard is closed")

		return (True,None)

		

#
# A local queue that compiles and runs the programs in the current node
#
class LocalSeqQueue(BasicQueue):


	# 
	# Checks the request and adds to the queue or marks as error
	#
	def check(self,request):

		logger.debug("Checking request: %d" % request.rid)

		# The next status is compiling
		request.status = Status.COMPILING

		# Set the timewall
		if not request.timewall:
			request.timewall = self.timewall
		else:
			request.timewall = min(self.timewall,request.timewall)

		# Insert the request into the database
		tablondb.insert_request(request)

		# Update the stats
		user = tablon.users.get(request.user)
		user.requests += 1
		user.save_stats()
		tablondb.update_user_stats(user)
		

	# Starts the compiling process
	def compile(self,request):

		logger.debug("Compiling request: %d" % request.rid)
		
		# Create the file
		code = request.src_name(self.SRCPATH)
		f = open(code,'w')
		f.write(request.content)
		f.close()


		if request.cl1 in ['ca','cg']:
			cl1 = 'CL1="-Xptxas -dlcm=%s"' % request.cl1
		else:
			cl1 = ""	

		# Compile
		cmd = ['make','-f',self.makefile,cl1,request.bin_name()]
		#print ' '.join(cmd)
		process = ExecutionThread(' '.join(cmd), cwd=self.SRCPATH,shell=True)
		out, err = process.communicate(fullout=True)

		logger.info("Request %d, user %s: compiling program: %s" % (request.rid,request.user,' '.join(cmd)))

		# Update the Status
		if process.returncode != 0:
			request.output = out
			request.status = Status.ERROR
			request.message = "Compile error"

			# Update the stats
			user = tablon.users.get(request.user)
			user.errors += 1
			user.save_stats()
			tablondb.update_user_stats(user)

			logger.info("Request %d, user %s: compiling error" % (request.rid,request.user))

		else:
			request.status = Status.DEPLOYING
			# Move the bin
			shutil.move('%s/%s' % (self.SRCPATH,request.bin_name()), self.BINPATH)

		# Update the database
		tablondb.update_request(request)


	# Execute the code
	def execute(self,request):

		# Do not allow simultaneous executions
		if self.exclusive:
			if BasicQueue.is_other_running:
				return
			else:
				BasicQueue.is_other_running = True

		# Leaderboard execution
		if request.leaderboard:

			# Get the leaderboard and phase
			lb = self.getLeaderboard(request.queue)
			phase = lb.getPhase(request.lb_phase)

			# Override the arguments if this is a leaderboard
			phase.override(request)

		logger.info("Request %d, user %s: executing %s" % (request.rid,request.user,request.program))

		# This method starts the execution with the appropriate command
		self.execute_run(request)
		
		# Update the database
		request.status = Status.RUNNING
		tablondb.update_request(request)



	# Execute the code
	# This is only the command call, so it can be modified by children classes
	def execute_run(self,request):

		# Execute the program
		cmd = ['./%s' % (request.bin_name())]
		if request.arguments:
			cmd += request.arguments.split()

		request.process = ExecutionThread(cmd, cwd=self.BINPATH)


	# Elapsed time since the executing start
	def elapsed(self,request):
		return request.process.elapsed()


	# Checks the finalization of the excution
	def execute_end(self,request):

		# Check if it is still running and the timewall
		if request.process.poll() == None:
			if self.elapsed(request) > request.timewall:
				self.kill_timewall(request)
			return

		out, err = request.process.communicate()
		request.output = out

		# Update the Status
		if request.process.returncode != 0:
			request.status = Status.ERROR
			if request.message == '':
				request.message = 'Execution error'
				request.time += self.elapsed(request)

			logger.info("Request %d, user %s: execution error %s" % (request.rid,request.user,request.program))

		else:
			request.status = Status.FINISHED
			request.time += self.elapsed(request)

			logger.info("Request %d, user %s: execution finished %s" % (request.rid,request.user,request.program))

		# Check the leaderboard
		self.execute_end_leaderboard(request)

		# Update the database
		tablondb.update_request(request)


		# Update the stats
		user = tablon.users.get(request.user)
		if request.status == Status.ERROR:
			user.errors += 1
		user.usedquota += self.elapsed(request)
		user.save_stats()
		tablondb.update_user_stats(user)


		logger.debug("Finished request (or not, lb!): %d" % request.rid)

		if self.exclusive:
			BasicQueue.is_other_running = False


	# Check the leaderboard when ending a request execution
	def execute_end_leaderboard(self,request):

		# If we are in a leaderboard execution
		if not request.leaderboard:
			return
		
		# Get the phase and check the result with it
		lb = self.getLeaderboard(request.queue)
		phase = lb.getPhase(request.lb_phase)

		# Continue to the next one
		if phase.check(request):

			request.status = Status.WAITING

			# Increase the leaderboard phase
			request.lb_phase += 1
			# Check if we have finished
			if request.lb_phase >= lb.numPhases():

				request.status = Status.FINISHED
				request.message = "Program passed"
				lb.add(request)



	# Kill a process if it has reached the timewall
	def kill_timewall(self,request):

		request.process.kill()
				
		request.message = "Timewall reached"
		request.status = Status.ERROR

		# Update the database
		tablondb.update_request(request)

		# Update the stats
		user = tablon.users.get(request.user)
		user.errors += 1
		user.usedquota += self.elapsed(request)
		user.save_stats()
		tablondb.update_user_stats(user)


		logger.info("Request %d, user %s: timewall %s" % (request.rid,request.user,request.program))

		if self.exclusive:
			BasicQueue.is_other_running = False


	def cancel(self,request):

		# Ignore the Queued request, they will not get into the database
		if request.status == Status.QUEUED:
			return

		# Ignore finished request
		if request.status == Status.FINISHED or request.status == Status.ERROR:
			return

		# Stop running request
		if request.status == Status.RUNNING:
			request.process.kill()

		request.status = Status.CANCELED
		request.message = 'Request canceled'
		tablondb.update_request(request)

		logger.debug("Request %d canceled" % (request.rid))


#
# A local queue for OpenMP programs
#
class LocalOpenMPQueue(LocalSeqQueue):


	# Execute the code setting OMP_NUM_THREADS 
	def execute_run(self,request):

		# ENV
		env = os.environ.copy()
		env['OMP_NUM_THREADS'] = str(request.threads)

		# Execute the program
		cmd = ['./%s' % (request.bin_name())]
		if request.arguments:
			cmd += request.arguments.split()

		request.process = ExecutionThread(cmd, env=env, cwd=self.BINPATH)





#
# A local queue for MPI programs
#
class LocalMPIQueue(LocalSeqQueue):


	# Execute the code using mpiexec
	def execute_run(self,request):

		# Execute the program
		cmd = ['mpiexec','-n',str(request.nprocs),'./%s' % (request.bin_name())]
		if request.arguments:
			cmd += request.arguments.split()

		request.process = ExecutionThread(cmd, cwd=self.BINPATH)


#
# A local hybrid queue for MPI+OpenMP programs
#
class LocalMPIOpenMPQueue(LocalSeqQueue):


	# Execute the code using mpiexec
	def execute_run(self,request):

		# ENV
		env = os.environ.copy()
		env['OMP_NUM_THREADS'] = str(request.threads)

		# Execute the program
		cmd = ['mpiexec','-n',str(request.nprocs),'./%s' % (request.bin_name())]
		if request.arguments:
			cmd += request.arguments.split()

		request.process = ExecutionThread(cmd, env=env, cwd=self.BINPATH)





#
# A local hybrid queue for MPI+OpenMP+CUDA programs
#
class SlurmQueue(LocalSeqQueue):


	def __init__(self,name,**kwargs):

		super(SlurmQueue, self).__init__(name,**kwargs)
		self.exclusive = False
		self.remote_binpath = kwargs.get('remote_binpath','/home/tablon/sandbox')
		self.slurm_queue = kwargs.get('slurm_queue','gpus')
		self.slurm_nodes = kwargs.get('slurm_nodes',None)
		self.slurm_usenodes = kwargs.get('slurm_usenodes',None)

	# Execute the code using mpiexec
	def execute_run(self,request):

		# ENV
		env = os.environ.copy()
		env['OMP_NUM_THREADS'] = str(request.threads)

		# Execute the program
		exe = request.bin_name()
		cmd = 'echo "sbcast -f %s/%s/%s %s/%s; srun %s/%s %s "'
		if request.arguments:
			args = request.arguments
		else:
			args = ''

		if self.slurm_usenodes:
			n = '-N '
		else:
			n = '-n '

		cmd = cmd % (os.getcwd(),self.BINPATH,exe,self.remote_binpath,exe,self.remote_binpath,exe,args)
		# ARTURO: GPUs
		cmd += ' | salloc -p %s %s %d --gres=gpu:1 ' % (self.slurm_queue,n,request.nprocs)
		if self.ngpus > 0:
			cmd += ' --gres=gpu:%d ' % (self.ngpus)

		# Add the nodes
		nodes = self.slurm_nodes
		if hasattr(request, 'slurm_nodes'):		
			nodes = requets.slurm_nodes
		if nodes != None:
			cmd += ' -w ' + ','.join(nodes)

		# Add the timewall
		m, s = divmod(request.timewall, 60)
		cmd += " -t 00:%02d:%02d " % (m,s)

		# ARTURO: TO CHECK THE COMMAND LINES
		#milog = open('/home/tablon/milog.txt','a');
		#milog.write('\n');
		#milog.write(cmd);
		#milog.write('\n');
		#milog.close();

		logger.info("alloc '%s'" % cmd)
		request.process = ExecutionThread(cmd, env=env, shell=True, cwd=self.remote_binpath, slurm=True)



	# Kill a process if it has reached the timewall
	def kill_timewall(self,request):
		# Let the slurm queues take care of the timewalls
		pass


