#!/usr/bin/python
# -*- coding: utf-8 -*-

################################################################################
#
# Servidor Tablon - LaunchServer
# Asignatura Computación paralela
# Javier Fresno, 2015
#
################################################################################


#
# Logging
#
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('tablon')

# Imports
import binascii, hashlib, pickle
import socket
import utils
import tablondb

# Import the execution requets object
import executionrequest

#
# Import authentication library
#
from load_srp import *

#
# Instance to tablon module
#
import sys
tablon = sys.modules['tablon'] 




#
# The LaunchServer class is the handler that manages the client requests (CONNECT)
#
class LaunchServer:


	# CONNECT request is used to send data with the client
	def do_CONNECT(self):

		# We have a new connection
		logger.debug("New connection from %s",self.client_address)

		# Authenticate the user
		if not self.authenticate():
			return

		# Manage the requst
		self.manage_request()

		return


	#
	# Authenticate the user
	#	
	def authenticate(self):

		# 1. Authentication
		try:

			# Get the version
			client_version = self.rfile.readline().strip()

			if(client_version != CONFIG.client_version):
				self.wfile.write('Unsuported client version, please download the client again.\n')
			else:
				# We send OK
				self.wfile.write('OK\n')

			# Get user name
			uname = self.rfile.readline().strip()
			A = binascii.a2b_hex(self.rfile.readline().strip())
			logger.debug("User %s connecting", uname)

			# Get the user in our list
			user = tablon.users.get(uname)
			if not user:
				raise Exception('User do not exists')
			self.user = user


			# Create and send challenge
			svr = srp.Verifier(uname, user.salt, user.vkey, A, hash_alg=srp.SHA256)
			s, B  = svr.get_challenge()

			if s is None or B is None:
				raise Exception('Challenge failed')
			self.wfile.write(binascii.b2a_hex(s)+'\n')
			self.wfile.write(binascii.b2a_hex(B)+'\n')

			# Get the solution to the challenge
			M = binascii.a2b_hex(self.rfile.readline().strip())

			# Verify session
			HAMK = svr.verify_session(M)

			if HAMK is None:
				raise Exception('Sesion not verified')

			# Send the session
			self.wfile.write(binascii.b2a_hex(HAMK)+'\n')
			if svr.authenticated():
				logger.info("User %s connected from %s", uname, self.client_address)
					
			# Get the session key
			key = svr.get_session_key()

			# Return
			self.username = uname
			self.userkey = key
			return True

		except Exception as e:
			user = tablondb.get_user_from_ip(self.client_address[0])
			logger.warning("Wrong connection from %s maybe '%s'", self.client_address, user)
			self.wfile.write('Wrong connection\n')

			print e
			import traceback
			print traceback.format_exc()


		return False
			

	#
	# Manage the request
	#	
	def manage_request(self):

		if not self.username:
			return

		try:

			# Get request len
			datalen = int(self.rfile.readline())
			if datalen > CONFIG.max_program_size:
				self.wfile.write('Request exceeds size limit')
				return

			# Check the restarttime
			if tablon.restartat != None:
				if (utils.get_timestamp() + CONFIG.time_before_restart) > tablon.restartat:
					self.wfile.write('Request not accepted, server is going to restart')
					return

			# Check quota
			if self.user.quota > 0 and not self.user.admin and self.user.quota < self.user.usedquota:
				if CONFIG.time_between_requests_no_quota == 0:
					self.wfile.write('Quota exhausted' )
					return
				else: 
					now = utils.get_timestamp()
					last = tablondb.last_timestamp(self.username)

					if(last+CONFIG.time_between_requests > now):
						self.wfile.write('Quota exhausted, only one request every %d seconds' % CONFIG.time_between_requests_no_quota )
						return


			# Check the time between
			if not self.user.admin:
				now = utils.get_timestamp()
				last = tablondb.last_timestamp(self.username)

				if(last+CONFIG.time_between_requests > now):
					self.wfile.write('Only one request every %d seconds' % CONFIG.time_between_requests )
					return


			# Read the data
			data = self.rfile.read(datalen)

			# Get the hash from the user
			usrhash = self.rfile.readline().strip()

			# Compute the hash of the file and the key
			m = hashlib.sha256()
			m.update(data)
			m.update(self.userkey)
		
			# Check hash
			if usrhash != m.hexdigest():
				error = "Error at send, hash do not match"
				logger.warning(error + ", user %s", self.username)
				self.wfile.write(error)
				return

			# Now its safe to read the data
			request = pickle.loads(data)

			# Sanitize the reques
			if not request.sanitize():
				self.wfile.write('Sanitize failed, wrong request')
				return	


			# Remove the reference from not admins
			if not self.user.admin:
				request.reference = False

			# Everything ok at this point
			self.wfile.write('Request sent successfully\n')
			logger.info("Request received successfully from user %s %s",self.username,self.client_address)


			# Set the queue to empty if it is None
			if not request.queue:
				request.queue = ''

			# Check if the queue exists
			if not request.queue in tablon.queues:
				self.wfile.write('Queue "%s" do not exits' % request.queue)
				return

			# Check if the queue support the number of threads
			queue =  tablon.queues[request.queue]		
			ok, msg = queue.check_nprocs_threads(request)
			if not ok:
				self.wfile.write(msg)
				return

			# Check extension
			ok, msg = queue.check_extension(request)
			if not ok:
				self.wfile.write(msg)
				return

			# Check the leaderboard (if any)
			ok, msg = queue.check_leaderboard(request)
			if not ok:
				self.wfile.write(msg)
				return

			# Check forbidden
			num,line = request.check_forbidden()
			if num:
				self.wfile.write('Forbidden code in Line %d: %s\n' % (num,line))
				return


			request.user = self.username
			request.address = self.client_address[0]
			request.addressname = self.address_string()


			# Init the request and adds all the missing fields
			request.init()

			# Finally add the request and give it a rid
			rid = tablon.exemanager.add(request)


			# Get the original port used to send the request
			if hasattr(CONFIG, 'real_port'):
				port = CONFIG.real_port
			else:
				try:
					port = utils.get_originalport(self.rfile._sock)
				except:
					port = CONFIG.port	


			# Check if port is 80 to skip the its number
			if port == 80:
				str_port = ''
			else:
				str_port = ':%d' % 	CONFIG.port	
	
			# Write the url 
			url = "http://%s%s/request?rid=%d" % (CONFIG.host,str_port,rid)
			self.wfile.write('Request id %d\n%s' % (rid,url) )


		except Exception as e:
			logger.warning("Error receiving request from %s: %s", self.username, self.client_address)
			self.wfile.write('Error receiving request\n')

			print e
			import traceback
			print traceback.format_exc()
















