#!/usr/bin/python
# -*- coding: utf-8 -*- 

################################################################################
#
# Servidor Tablon - DataBase
# Asignatura Computación paralela
# Javier Fresno, 2016
#
################################################################################

#
# Logging
#
import logging
logger = logging.getLogger('tablon')

# Imports
import MySQLdb
import threading
import time
import functools
import os
import utils
from distutils.version import LooseVersion as Version
from datetime import datetime, timedelta 




#
# Reference to this module
#
import sys
tablondb = sys.modules[__name__]
tablondb.initiated = False

#
# The database connections
#
connections = dict()

#
# Create the connection for the current thread
#
def get_connection():

	thread = threading.current_thread()
	ident = thread.ident

	if not ident in connections:
		logger.debug("New BD connection for thread: " + str(ident))
		conn = MySQLdb.connect(CONFIG.bd_host,CONFIG.bd_user,CONFIG.bd_pass,CONFIG.bd_name,charset='utf8',use_unicode=True)
		connections[ident] = conn

	return connections[ident]




#
# Close all the database connections
# TODO: bug other threads can't close this
#
def close():

	return

	logger.info("Closing database connections %d" % len(connections) )

	for ident in connections:
		connections[ident].close()




#
# This is a decorator class
# This replaces the functions that use the dabase to obtain a DB cursor
# and do an optional commit at the end
# It also wrappers the execution in a try.
# The default value es returned if there is any exception
#
class dbcursor(object):

	# Decorator constructor
	def __init__(self,commit=False, default=None):

		# Commit flag and default value
		self.commit = commit
		self.default = default

	# Call the decorator
	def __call__(self, func=None):

		# If we have no function, this means the user has not properly
		# used the dbcursor decorator
		if func == None:
			raise Exception("You forgot the () in the decorator @dbcursor(), you need to init and call it!!! ")		

		# Reference to the decorator
		decorator = self

		# The returned function (javascript style)
		# This other decorator is used to mark the function
		# as a wraper for the debuger and trace
		@functools.wraps(func)
		def dbcursor_and_call(*args, **kwargs):

			logger.debug("Database call. Call to " + func.__name__)

			# If it is not initiated and it isn't a init call
			# it can be an error if we do not define a default value
			if not tablondb.initiated and func.__name__ != 'init':
				if decorator.default != None:
					return decorator.default
				raise Exception("TablondDB has not been initiated func '%s'" % func.__name__)


			# All is in a big try
			try:

				# This variable is True when we are the parent database call
				parentquery = False

				# Connect and get cursor
				if not 'cursor' in kwargs:
					DB = get_connection()
					cursor = DB.cursor(MySQLdb.cursors.DictCursor)
					parentquery = True


				else:
					cursor = kwargs['cursor']
					del kwargs['cursor']


				# Call the orginal function
				res = func(cursor, *args, **kwargs)

				# Save the database
				if parentquery:
					if decorator.commit:
						DB.commit()
					else:
						# This is needed to update the DB conection
						DB.rollback()
					cursor.close()
				else:
					# If we are not a parent query, we cannot close the cursor
					pass

				# Return value
				return res

			except Exception as e:

				if decorator.default:
					import traceback
					logger.error("1 - error " + str(e) + "\n" + traceback.format_exc())
					return decorator.default

				logger.error("Error in database. Call to " + func.__name__)

				import traceback
				logger.error("2 - error " + str(e) + "\n" + traceback.format_exc())

				raise Exception("Database exception")


		return dbcursor_and_call




#
# Check if table exists
#
@dbcursor()
def check_table_exists(cursor,table):

	cursor.execute("SELECT * FROM information_schema.tables WHERE table_name = %s",(table,))

	if cursor.fetchone():
		return True
	else:
		return False



#
# Return the db version or None if it is empty
#
@dbcursor()
def get_dbversion(cursor):

	# Return the database version for 5.4+
	try:
		cursor.execute('SELECT version FROM tablondb')
		return Version(cursor.fetchone().values()[0])
	except:
		pass

	# Check if the database is empty
	cursor.execute('SELECT COUNT(*) FROM information_schema.tables WHERE table_schema = "tablon";')
	if int(cursor.fetchone().values()[0]) < 6:
		return None

	raise Exception("Version of database undefined")


@dbcursor()
def get_last_request(cursor):

	last = 0
	if hasattr(CONFIG,"last_request"):
		last = CONFIG.last_request
	try:
		cursor.execute('SELECT MAX(rid) FROM requests')
		return max(int(cursor.fetchone().values()[0]),last)
	except:
		return last



#
# Init the database and create the tables if needed
#
@dbcursor(commit=True)
def init(cursor,userlist,leaderboards):

	logger.info("Init database ...")
	tablondb.initiated = True


	# Check version
	# ----------------------------------------------- #
	dbversion = get_dbversion(cursor=cursor)
	logger.info("Database file version: %s" % dbversion)


	# Add new version
	# ----------------------------------------------- #

	# Create the table
	if not check_table_exists("tablondb",cursor=cursor):
		cursor.execute('CREATE TABLE IF NOT EXISTS tablondb(id INTEGER PRIMARY KEY, version VARCHAR(20))')
	cursor.execute('INSERT INTO tablondb (id,version) VALUES (1,%s) ON DUPLICATE KEY UPDATE version=VALUES(version)',(CONFIG.version,))



	# User table
	# ----------------------------------------------- #

	# Create the table	
	if not check_table_exists("users",cursor=cursor):
		cursor.execute('''
			CREATE TABLE users(
				name VARCHAR(255) PRIMARY KEY, quota INTEGER, usedquota FLOAT DEFAULT 0,
				requests INTEGER DEFAULT 0, errors INTEGER DEFAULT 0, admin INTEGER DEFAULT 0)
			DEFAULT CHARSET=utf8
		''')

	# Remove db user that do not exists on the list
	cursor.execute('SELECT name FROM users')
	rows = cursor.fetchall()
	toremove = []
	for row in rows:
		name = row['name']
		if(not userlist.get(name)):
			toremove.append(name)

	# Remove the users
	for name in toremove:
		logger.debug('Removing user %s from db' % name)
		cursor.execute('DELETE FROM users WHERE name = %s',(name,))
			
	# Add the new users
	for user in userlist:
		cursor.execute('INSERT INTO users(name, quota, admin) VALUES(%s, %s, %s) ON DUPLICATE KEY UPDATE quota=VALUES(quota),admin=VALUES(admin)',
			(user.name,user.quota,user.admin))
		update_user_stats(user,cursor=cursor)


	# Statuses table
	# ----------------------------------------------- #

	# Create the table
	if not check_table_exists("statuses",cursor=cursor):
		cursor.execute('''
			CREATE TABLE statuses
				(id INTEGER PRIMARY KEY, enum VARCHAR(255))
		''')

	values = [(1,"queued"),(2,"compiling"),(3,"deploying"),
			(4,"waiting"),(5,"running"),(6,"finished"),
			(7,"canceled"),(9,"error")]
	cursor.executemany('INSERT IGNORE INTO statuses VALUES(%s,%s)',values)


	# Request table
	# ----------------------------------------------- #

	# Create the table
	if not check_table_exists("requests",cursor=cursor):
		cursor.execute('''
			CREATE TABLE requests(rid INTEGER PRIMARY KEY,
			user VARCHAR(255) REFERENCES users(name), program VARCHAR(255),
			arguments TEXT,
			queue VARCHAR(255), nprocs INT, threads INT, time REAL, leaderboard INT,
			output TEXT, status INT REFERENCES statuses(id), message TEXT,
			timestamp INT, address TEXT, addressname TEXT,
			hash TEXT, clines INT, tokens INT, complexity INT)
			DEFAULT CHARSET=utf8
		''')

	# Not FINISHED -> ERROR
	cursor.execute('UPDATE requests SET status=9, message="Server crash" WHERE status<6')


	# Leaderboards table
	# ----------------------------------------------- #
	if not check_table_exists("leaderboards",cursor=cursor):
		cursor.execute('''
			CREATE TABLE IF NOT EXISTS leaderboards(rid INTEGER PRIMARY KEY,
			user VARCHAR(255) REFERENCES users(name), program VARCHAR(255),
			queue VARCHAR(255), time REAL, rtime REAL, timestamp INT, intxt INT)
			DEFAULT CHARSET=utf8
		''')
	if not check_table_exists("leaderboardsbackup",cursor=cursor):
		cursor.execute('''
			CREATE TABLE IF NOT EXISTS leaderboardsbackup(rid INTEGER PRIMARY KEY,
			user VARCHAR(255) REFERENCES users(name), program VARCHAR(255),
			queue VARCHAR(255), time REAL, rtime REAL, timestamp INT, intxt INT)
			DEFAULT CHARSET=utf8
		''')

	# Mark all the items as NOT in the txt
#	cursor.execute('UPDATE leaderboards SET intxt=0')
#
#	# Add leaderboards from files
#	for lb in leaderboards:
#		for item in lb:
#
#			# Using the magic of the decorator we can pass
#			# our current cursor
#			# We update the item in the table and set the intxt flag
#			add_leaderboard(item,cursor=cursor)
#
#	# Make a backup and remove the items that are no in the txt
#	cursor.execute('INSERT IGNORE INTO leaderboardsbackup (SELECT * FROM leaderboards WHERE intxt=0) ')
#	cursor.execute('DELETE FROM leaderboards WHERE intxt=0')

	# All done
	# ----------------------------------------------- #
	logger.info("Database DONE")





#
# Used quota of an user
#
@dbcursor(default=0)
def usedquota(cursor,name):

	# Get the quota
	cursor.execute('SELECT usedquota FROM users WHERE name = %s',(name,))
	row = cursor.fetchone()
	return int(row['usedquota'])



#
# Timestamp of the last request of one user
#
@dbcursor(default=0)
def last_timestamp(cursor,name):

	# Get the quota
	cursor.execute('SELECT  COALESCE(MAX(timestamp),0) AS timestamp FROM requests WHERE user = %s',(name,))
	row = cursor.fetchone()
	return int(row['timestamp'])



#
# Update the user stats: quota and errors
# We can pass a cursor to this function to speedup it when
# it is called in the init
#
@dbcursor(commit=True)
def update_user_stats(cursor,user):

	# Insert the request
	cursor.execute('UPDATE users SET usedquota=%s, requests=%s, errors=%s WHERE name=%s',
		(user.usedquota,user.requests,user.errors,user.name))






# Update the status, message, time, and output
@dbcursor(commit=True)
def update_request(cursor,r):

	# Insert the request
	values = (r.time,r.status,r.message,r.output,r.rid)
	cursor.execute('UPDATE requests SET time=%s, status=%s, message=%s, output=%s WHERE rid=%s',values)




# Insert a new request
@dbcursor(commit=True)
def insert_request(cursor,r):

	# Insert the request
	values = (r.rid,r.user,r.program,r.arguments,r.queue,r.nprocs,r.threads,r.time,
	int(r.leaderboard),r.output,r.status,r.message,r.timestamp,
	r.address,r.addressname,
	r.hash,r.lines,r.tokens,r.complexity)

	cursor.execute('INSERT INTO requests VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s);',values)



@dbcursor()
def show_request_list(cursor,last=0):

	cursor.execute('''
		SELECT	rid,user,queue,nprocs,threads,program,status,
				message,time,leaderboard,statuses.enum AS statusname,timestamp
		FROM requests
		INNER JOIN statuses ON
			requests.status = statuses.id
		WHERE rid > %s
		ORDER BY rid DESC
		LIMIT %s
	''',(last,CONFIG.requestnumber))

	return cursor.fetchall()



@dbcursor()
def show_request(cursor,rid):
	
	cursor.execute('''
		SELECT	r.*, s.enum AS statusname, l.time AS ltime
		FROM requests r
		LEFT OUTER  JOIN leaderboards l ON
			r.rid = l.rid
		INNER  JOIN statuses s ON
			r.status = s.id
		WHERE r.rid = %s
	''',(rid,))

	return  cursor.fetchone()



@dbcursor(commit=True)
def add_leaderboard(cursor,item):

	# Insert the new item
	cursor.execute('''
		INSERT INTO leaderboards (rid, user, program, queue, time, rtime, timestamp, intxt)
		VALUES(%s,%s,%s,%s,%s,%s,%s,1)
		ON DUPLICATE KEY UPDATE user=VALUES(user),program=VALUES(program),
		queue=VALUES(queue),time=VALUES(time),rtime=VALUES(rtime),
		timestamp=VALUES(timestamp),intxt=VALUES(intxt)
	''',(item.rid,item.user,item.program,item.queue,item.time,item.rtime,item.timestamp))







@dbcursor()
def show_leaderboard(cursor,name):
	
	# Delete the temp table if exists
	cursor.execute('DROP TABLE IF EXISTS templeaderboard')

	# Init the row variable
	cursor.execute('''
		set @row = 0; 
	''')


	# Create a new table with this leaderboard,
	# insert first the ordered users

	cursor.execute('''
		CREATE TEMPORARY TABLE templeaderboard as 
		SELECT @row := @row + 1 AS pos, lb2.*
		FROM(
			SELECT user, MIN(time) as mintime
			FROM leaderboards
			WHERE queue = %s AND user <> 'reference'
			GROUP BY user
		) lb1
		INNER JOIN leaderboards lb2 
		ON lb1.user = lb2.user AND lb1.mintime = lb2.time
		ORDER BY lb2.time;

	''',(name,))


	# insert first the reference values
	cursor.execute('''
		INSERT INTO templeaderboard (pos,rid,user,program, queue, time, rtime, timestamp)
		SELECT 0,rid,user,program, queue, time, rtime, timestamp FROM leaderboards
		WHERE queue = %s AND user = 'reference';
	''',(name,))


	# Return the table and do a final order
	cursor.execute('SELECT pos, rid, user, program, queue, time, rtime, timestamp FROM templeaderboard ORDER BY time;')


	return cursor.fetchall()





@dbcursor()
def show_leaderboard_admin(cursor,name):

	# Delete the temp table if exists
	cursor.execute('DROP TABLE IF EXISTS templeaderboard2')

	# Create a new table with this leaderboard,
	# insert first the ordered users.

	# Init the row variable
	cursor.execute('''
		set @row = 0; 
	''')

	# This uses a join and is super fast :)
	cursor.execute('''
		CREATE TEMPORARY TABLE templeaderboard2 as 
		SELECT @row := @row + 1 AS pos, lb2.*
		FROM(
			SELECT user, MIN(time) as mintime
			FROM leaderboards
			WHERE queue = %s AND user <> 'reference' 
			GROUP BY user
		) lb1
		INNER JOIN leaderboards lb2 
		ON lb1.user = lb2.user AND lb1.mintime = lb2.time
		ORDER BY lb2.time;
	''',(name,))


	cursor.execute('''
		SELECT name as user2, tlb.*
		FROM users u
		LEFT OUTER JOIN templeaderboard2 tlb
		ON u.name = tlb.user
		WHERE admin = 0
		ORDER BY u.name;
	''')



	return cursor.fetchall()





@dbcursor()
def show_leaderboard_admin_user(cursor,name,user):
	

	cursor.execute('''
		SELECT r2.firstrid AS same, l.*, r.hash
		FROM (
			SELECT * FROM leaderboards
			WHERE queue = %s AND user = %s
			ORDER BY time
		) l
		INNER JOIN requests r ON r.rid = l.rid
		LEFT JOIN (
			SELECT MIN(rid) as firstrid, hash from requests
			GROUP BY hash
		) r2
		ON r.hash = r2.hash''',(name,user))


	return cursor.fetchall()







@dbcursor()
def show_users(cursor):
	
	cursor.execute('SELECT * FROM users ORDER BY admin, name',)

	return cursor.fetchall()







@dbcursor(default=None)
def get_user_from_ip(cursor,address,seconds=60*15):


	cursor.execute("""
		SELECT user
		FROM requests
		WHERE address=%s AND timestamp > current_timestamp() - %s
		GROUP BY user ORDER BY timestamp DESC
		LIMIT 1""",(address,seconds))

	row = cursor.fetchone()
	if row != None:
		return row['user']
	else:
		return None















# Functions for stats
# - # - # - # - # - # - # - # - # - # - # - # - # - # - # - # - # - # - # - # 






@dbcursor()
def show_stats_request_status(cursor,ini,end):
	
	cursor.execute('''
		SELECT statuses.enum AS name, COUNT(*) AS number
		FROM requests
		INNER JOIN statuses ON requests.status = statuses.id
		WHERE timestamp > %s AND timestamp < %s
		GROUP BY status
	''',(ini,end))

	return cursor.fetchall()

@dbcursor()
def show_stats_error_types(cursor,ini,end):
	
	cursor.execute('''
		SELECT COUNT(rid) AS number, message
		FROM requests
		WHERE status=9 AND timestamp > %s AND timestamp < %s
		GROUP BY message
	''',(ini,end))

	return cursor.fetchall()




@dbcursor()
def show_stats_total_requests(cursor,ini,end):

	cursor.execute('''
		SELECT COUNT(*) AS number, COALESCE(SUM(time),0) AS time
		FROM requests
		WHERE timestamp > %s AND timestamp < %s
	''',(ini,end))

	return cursor.fetchone()


@dbcursor()
def show_stats_first_after_ts(cursor,ts):

	cursor.execute('''
		SELECT MIN(timestamp) AS first
		FROM requests
		WHERE timestamp > %s
	''',(ts,))

	return cursor.fetchone().values()[0]


@dbcursor()
def show_stats_last_before_ts(cursor,ts):

	cursor.execute('''
		SELECT MAX(timestamp) AS first
		FROM requests
		WHERE timestamp < %s
	''',(ts,))

	return cursor.fetchone().values()[0]



# Number of requests each day
@dbcursor()
def show_stats_request_byday(cursor,ini,end):

	cursor.execute('''
		SELECT date(FROM_UNIXTIME(timestamp)) AS date, COUNT(*) AS number
		FROM requests
		WHERE timestamp > %s AND timestamp < %s
		GROUP BY date(FROM_UNIXTIME(timestamp))
	''',(ini,end))

	return cursor.fetchall()



@dbcursor()
def show_stats_request_byhour(cursor,ini,end):

	cursor.execute('''

		SELECT hour, MAX(number) as number
		FROM (
			SELECT HOUR(FROM_UNIXTIME(timestamp)) AS hour, COUNT(*) AS number
					FROM requests
					WHERE timestamp > %s AND timestamp < %s
					GROUP BY HOUR(FROM_UNIXTIME(timestamp))
			UNION
			SELECT '00' as hour, 0 as number 
			UNION
			SELECT '01' as hour, 0 as number
			UNION
			SELECT '02' as hour, 0 as number
			UNION
			SELECT '03' as hour, 0 as number
			UNION
			SELECT '04' as hour, 0 as number
			UNION
			SELECT '05' as hour, 0 as number
			UNION
			SELECT '06' as hour, 0 as number
			UNION
			SELECT '07' as hour, 0 as number
			UNION
			SELECT '08' as hour, 0 as number
			UNION
			SELECT '09' as hour, 0 as number
			UNION
			SELECT '10' as hour, 0 as number
			UNION
			SELECT '11' as hour, 0 as number
			UNION
			SELECT '12' as hour, 0 as number
			UNION
			SELECT '13' as hour, 0 as number
			UNION
			SELECT '14' as hour, 0 as number
			UNION
			SELECT '15' as hour, 0 as number
			UNION
			SELECT '16' as hour, 0 as number
			UNION
			SELECT '17' as hour, 0 as number
			UNION
			SELECT '18' as hour, 0 as number
			UNION
			SELECT '18' as hour, 0 as number
			UNION
			SELECT '19' as hour, 0 as number
			UNION
			SELECT '20' as hour, 0 as number
			UNION
			SELECT '21' as hour, 0 as number
			UNION
			SELECT '22' as hour, 0 as number
			UNION
			SELECT '23' as hour, 0 as number
		) AS T
		GROUP BY hour
		ORDER BY hour

	''',(ini,end))


	return cursor.fetchall()



@dbcursor()
def show_stats_request_byweekday(cursor,ini,end):

	cursor.execute('''

		SELECT day, MAX(number) as number
		FROM (
			SELECT DAYOFWEEK(FROM_UNIXTIME(timestamp)) AS day, COUNT(*) AS number
					FROM requests
					WHERE timestamp > %s AND timestamp < %s
					GROUP BY DAYOFWEEK(FROM_UNIXTIME(timestamp))
			UNION
			SELECT '0' as day, 0 as number
			UNION
			SELECT '1' as day, 0 as number
			UNION
			SELECT '2' as day, 0 as number
			UNION
			SELECT '3' as day, 0 as number
			UNION
			SELECT '4' as day, 0 as number
			UNION
			SELECT '5' as day, 0 as number
			UNION
			SELECT '6' as day, 0 as number
		) AS T
		GROUP BY day
		ORDER BY day

	''',(ini,end))


	return cursor.fetchall()





@dbcursor()
def show_stats_leaderboard_plot(cursor,lbname):

	"""
	This is the big function to create the leaderboard plot
	"""

	import datetime


	# Get the min timestamp
	cursor.execute('''	SELECT min(timestamp) FROM leaderboards
						WHERE queue = %s''',(lbname,))
	mintimestamp = cursor.fetchone().values()[0]
	if mintimestamp == None:
		result = {}
		result["nadie"] = [(datetime.datetime(2016, 1, 1, 1, 1), 1)]
		return result

	# Get the max timestamp
	cursor.execute('''SELECT	max(timestamp) FROM leaderboards
								WHERE queue = %s''',(lbname,))
	maxtimestamp = cursor.fetchone().values()[0]


	# Seconds in a day
	dayseconds = 86400


	# Get the min and max days
	mintimestamp = int((mintimestamp/dayseconds)*dayseconds)
	maxtimestamp = int((maxtimestamp/dayseconds)*dayseconds)


	# Create the temp table

	if check_table_exists("temp_dominating1",cursor=cursor):
		cursor.execute('DROP TABLE IF EXISTS temp_dominating1')
	cursor.execute('CREATE TEMPORARY TABLE temp_dominating1 (user TEXT, pos INT, dayts INT)')


	# Iterate for each day
	for ts in range(mintimestamp,maxtimestamp+dayseconds,dayseconds):

		# Delete the second temp table
		cursor.execute('DROP TABLE IF EXISTS temp_dominating2')

		# Create a new table with this leaderboard,
		# insert first the ordered users.
		# This way we can use the ROWID as pos.

		# This uses a join and is super fast :)
		cursor.execute('''
			CREATE TEMPORARY TABLE temp_dominating2 as 
			SELECT lb2.*
			FROM(
				SELECT user, MIN(time) as mintime
				FROM leaderboards
				WHERE queue = %s AND user <> 'reference' AND timestamp < %s 
				GROUP BY user
			) lb1
			INNER JOIN leaderboards lb2 
			ON lb1.user = lb2.user AND lb1.mintime = lb2.time
			ORDER BY lb2.time;
		''',(lbname,ts+dayseconds))

		# Init the row variable
		cursor.execute('''
			set @row = 0; 
		''')


		# Add temp2 into temp1 with the rowid as pos
		cursor.execute('''
			INSERT INTO temp_dominating1
			SELECT user, @row := @row + 1 AS pos, %s
			FROM temp_dominating2
		''',(ts,))



	# Get the max pos, total number of tries
	cursor.execute('SELECT max(pos) as maxpos FROM temp_dominating1')
	maxpos = int(cursor.fetchone()['maxpos'])


	# Get the list of users
	cursor.execute('SELECT user FROM temp_dominating1 GROUP BY user')
	rows = cursor.fetchall()

	users = []
	for row in rows:
		users.append(row["user"])



	# Result
	# The result is dict with users as keys
	# each user has a list with tuples (day, pos)
	result = {}


	# Iterate for the users
	for user in users:

		# Get the dates for one user
		cursor.execute('SELECT * FROM temp_dominating1 WHERE user = %s',(user,))	
		rows = cursor.fetchall()

		# Compose the result for the user
		result[user] = []
		for row in rows:
			date = datetime.datetime.fromtimestamp(row['dayts'])
			pos = int(row['pos'])
			result[user].append((date,pos))



	# Return the result
	return result





