#!/usr/bin/python
# -*- coding: utf-8 -*- 

################################################################################
#
# Servidor Tablon - ThreadPoolMixIn
# Asignatura Computación paralela
# Javier Fresno, 2015
#
################################################################################

# This is like a ThreadMixIn but with a pool of threads
# Based on:
# http://code.activestate.com/recipes/574454-thread-pool-mixin-class-for-use-with-socketservert/


from SocketServer import ThreadingMixIn, BaseServer
from Queue import Queue
import threading, socket


#
# Class that uses a thread pool instead of a new thread on every request
#
class ThreadPoolMixIn(ThreadingMixIn):

	numThreads = 1
	allow_reuse_address = True  # seems to fix socket.error on server restart

	# Handle request
	def serve_forever(self):

		# set up the threadpool
		self.requests = Queue(self.numThreads)

		# Flag to exit
		self.keep_running_pool = True

		# Create a list of threads
		threads = []
		for x in range(self.numThreads):
			t = threading.Thread(target = self.process_request_thread)
			t.start()
			threads.append(t)

		# server main loop
		while self.keep_running_pool:
			self.handle_request()

		self.server_close()

		# Add an empty request to stop the threads
		for t in threads:
			self.requests.put((None,None))

		# Wait for the threads to exit
		for t in threads:
			t.join()



	# obtain request from queue instead of directly from server socket
	def process_request_thread(self):
		while self.keep_running_pool:
			request, address = self.requests.get()
			if request != None and  address != None:
				ThreadingMixIn.process_request_thread(self, request, address)

	# Get a requets and add it to the queue
	def handle_request(self):

		try:
			request, client_address = self.get_request()
		except socket.error:
			return

		if self.verify_request(request, client_address):
			self.requests.put((request, client_address))

	# Shutdown the server
	def shutdown(self):
		self.keep_running_pool = False
		# The fake request breaks the serve_forever loop
		self.fake_http_request()


	# Fake request used to stop the server
	def fake_http_request(self):
		sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
		sock.connect(self.server_address)
		sock.close()


