#!/usr/bin/python
# -*- coding: utf-8 -*-

################################################################################
#
# Servidor Tablon - Utils
# Asignatura Computación paralela
# Javier Fresno, 2015
#
################################################################################

import datetime
import socket
import struct
import lizard
import hashlib

#
# This function obtains the original port used to connect from a socket.
# I used it to know if the user connect to 80 or 8080
#
def get_originalport(thesocket):

	# Get the original port used to send the request
	# This calls the low level getsockopt function
	try:
		SO_ORIGINAL_DST = 80
		BUFFER_SIZE = 16
		dst_info = thesocket.getsockopt(socket.SOL_IP, SO_ORIGINAL_DST, BUFFER_SIZE)
		dst_port, dst_ip = struct.unpack("!2xH4s8x", dst_info)
		#dst_srt_ip = socket.inet_ntoa(dst_ip)
		return dst_port
	except:
		return None




#
# Get the UNIX timestamp
# I use this because it simpler when using dates in the browser with javascript
#
def get_timestamp():
	return ((datetime.datetime.utcnow())-datetime.datetime(1970,1,1)).total_seconds()


def get_timestamp_day(year=None,month=None,day=None):

	if year == None:
		day = datetime.date.today()
	else:
		day = datetime.date(year,month,day)

	day = datetime.datetime.combine(day, datetime.datetime.min.time())

	return (day-datetime.datetime(1970,1,1)).total_seconds()


#
# Print a timestamp
#
def print_timestamp(timestamp):

	if timestamp == None:
		return "None"

	dtime = datetime.datetime.fromtimestamp(int(timestamp))
	return dtime.strftime("%d-%m-%Y %H:%M")



#
# Jinja filte to cut long string
#
def text_overflow(text,size=10):
	if(len(text) > size+3):
		return text[0:size] + "..."
	else:
		return text   


#
# Jinja filter to print elapsed times given seconds
#
def print_elapsed(seconds,precision=3):

	# Get the hours, minutes and seconds
	minutes = int(seconds / 60)
	seconds = seconds % 60
	hours = int(minutes / 60)
	minutes = minutes % 60

	# The string
	output = ''

	# We have hours
	if hours > 0:
		output += '%dh ' % hours

		# If we have hours, round the minutes and discard seconds
		minutes = round(minutes)
		seconds = 0

	# We have minutes
	if minutes > 0:
		output += '%dm ' % minutes

		# If we have minutes, round the seconds
		seconds = round(seconds)
		if seconds > 0 :
			output += '%ds' % (seconds)

	# Only seconds
	else:
		if seconds > 0 or output == '':
			output += '%0.*fs' % (precision,seconds)


	return output


#
# RGB tuple to html color
#
def RGBToHTMLColor(rgb_tuple):
    """ convert an (R, G, B) tuple to #RRGGBB """
    hexcolor = '#%02x%02x%02x' % rgb_tuple
    # that's it! '%02x' means zero-padded, 2-digit hex values
    return hexcolor

#
# html color to RGB
#
def HTMLColorToRGB(colorstring):
    """ convert #RRGGBB to an (R, G, B) tuple """
    colorstring = colorstring.strip()
    if colorstring[0] == '#': colorstring = colorstring[1:]
    if len(colorstring) != 6:
        raise ValueError, "input #%s is not in #RRGGBB format" % colorstring
    r, g, b = colorstring[:2], colorstring[2:4], colorstring[4:]
    r, g, b = [int(n, 16) for n in (r, g, b)]
    return (r, g, b)


#
# Degradate from one color to another
# Used for the quotas in the user tab
#
def color_percentage(value,minvalue=0.8,maxvalue=1.0,inicolor="#707070",endcolor="#9C0000"):

	if value < minvalue:
		return inicolor

	if value > maxvalue:
		return endcolor

	pos1 = float(value-minvalue) / float(maxvalue-minvalue)
	pos2 = float(maxvalue-value)  / float(maxvalue-minvalue)


	color1 = HTMLColorToRGB(inicolor)
	color2 = HTMLColorToRGB(endcolor)

	newcolor = [int(color1[i]*(1.0-pos1) + color2[i]*(1.0-pos2))  for i in range(3)] 

	return RGBToHTMLColor(tuple(newcolor))



#
# Function to compress a python script as an self-extraction python script
# https://github.com/liftoff/pyminifier
#
def gz_pack(source):
	"""
	Returns 'source' as a gzip-compressed, self-extracting python script.
	"""
	import zlib, base64, re
	shebang = re.compile('^#\!.*$')
	out = ""
	# Preserve shebangs (don't care about encodings for this)
	first_line = source.split('\n')[0]
	if shebang.match(first_line):
		out = first_line + '\n'
	compressed_source = zlib.compress(source.encode('utf-8'))
	out += 'import zlib, base64\n'
	out += "exec(zlib.decompress(base64.b64decode('"
	out += base64.b64encode(compressed_source).decode('utf-8')
	out += "')))\n"
	return out



def code_hash(content):

	import hashlib
	hasher = hashlib.sha1()
	hasher.update(content)
	return hasher.hexdigest()


def code_hash_file(fname):

	try:

		with open(fname, 'rb') as f:
			f.readline()
			f.readline()
			f.readline()
			content = f.read()
			return code_hash(content)
	except:
		pass

	return None





def code_stats(fname,content):

	info = lizard.analyze_file.analyze_source_code(fname,content)

	if info == None:
		return(0,0,0)
	nloc = info.nloc
	ccn = 0
	ntok = 0

	for f in info.function_list:
		ccn += f.cyclomatic_complexity
		ntok += f.token_count

	return (nloc,ccn,ntok)



def code_stats_file(fname):

	info = lizard.analyze_file(fname)
	if info == None:
		return(0,0,0)
	nloc = info.nloc
	ccn = 0
	ntok = 0

	for f in info.function_list:
		ccn += f.cyclomatic_complexity
		ntok += f.token_count

	return (nloc,ccn,ntok)



#
# Cache decorator
#
class cached(object):

	def __init__(self, max_age=0):
		self.cache = {}
		self.max_age = max_age


	# Call the decorator
	def __call__(self, func=None):

		# If we have no function, this means the user has not properly used the decorator
		if func == None:
			raise Exception("You forgot the () in the decorator, you need to init and call it!!! ")		

		# The returned function (javascript style)
		# This other decorator is used to mark the function
		# as a wraper for the debuger and trace
		@functools.wraps(func)
		def inner(*args, **kwargs):

			key = (func,args)
			#print key
			#print (key in self.cache)

			#if key in self.cache:
			#	print (datetime.now() - self.cache[key]['fetch_time'])

			if not self.max_age or key not in self.cache or (datetime.now() - self.cache[key]['fetch_time']) > self.max_age:
				data = func(*args, **kwargs)
				self.cache[key] = {'data': data, 'fetch_time': datetime.now()}
			else:
				logger.info("Cached call. Call to " + func.__name__)

			return self.cache[key]['data']


		return inner






















