#!/usr/bin/python
# -*- coding: utf-8 -*-

################################################################################
#
# Servidor Tablon - WebServer
# Asignatura Computación paralela
# Javier Fresno, 2015
#
################################################################################


#
# Logging
#
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('tablon')


# Other imports
import re, os 
from datetime import timedelta, datetime
import tablondb
import utils
import urlparse
import functools
import threading
import cgi
import Cookie
import urlparse
import time


#
# jinja2 template module
#
from jinja2 import Environment, FileSystemLoader
env = Environment(loader=FileSystemLoader('templates'))


env.filters['print_timestamp'] = utils.print_timestamp
env.filters['text_overflow'] = utils.text_overflow
env.filters['print_elapsed'] = utils.print_elapsed
env.filters['color_percentage'] = utils.color_percentage


#
# Instance to the tablon module
#
import sys
tablon = sys.modules['tablon'] 

#
# Class context with the default values for the jinja2 templates
#
class Context(dict):

	# Constructor
	def __init__(self):

		# The navigation menu
		menu = []
		menu.append({'name':'Home','url':'/'})
		menu.append({'name':'Users','url':'/users'})
		menu.append({'name':'Queues','url':'/queues'})
		menu.append({'name':'Leaderboards','url':'/leaderboards'})
		menu.append({'name':'FAQ','url':'/faq'})
		menu.append({'name':'Stats','url':'/stats'})
		self['menu'] = menu
		self['version'] = CONFIG.version
		self['hostname'] = CONFIG.hostname

		if hasattr(CONFIG, 'favicon'):
			self['favicon'] = CONFIG.favicon


class AdminContext(dict):

	# Constructor
	def __init__(self):

		# The navigation menu
		menu = []
		menu.append({'name':'Admin','url':'/admin'})
		self['menu'] = menu
		self['version'] = CONFIG.version
		self['hostname'] = CONFIG.hostname

		if hasattr(CONFIG, 'favicon'):
			self['favicon'] = CONFIG.favicon


class WebPageCacheItem:


	def __init__(self,cache):

		self.cache = cache
		self.time = utils.get_timestamp()
		self.cache_content_type = webrequest.content_type
		self.cache_response = webrequest.response
		self.cache_cachecontrol = webrequest.cachecontrol

	def get(self):
		webrequest.content_type = self.cache_content_type
		webrequest.response = self.cache_response 
		webrequest.cachecontrol = self.cache_cachecontrol
		return self.cache


class WebPageCache:

	def __init__(self,cachetime):

		self.cache = dict()
		self.cachetime = cachetime

	def get(self):

		# Remove old caches
		for path, item in self.cache.items():
			if (self.cachetime + item.time) < utils.get_timestamp():
				del self.cache[path]

		# Return the selected cache
		for path, item in self.cache.iteritems():
			if path == webrequest.path:
				return item.get()
		
		# There is no cache
		return None


	def set(self,cache):
		self.cache[webrequest.path] = WebPageCacheItem(cache)

		



class WebPage:

	def __init__(self,path,func,cache=None):

		logger.info("Created page '%s'" % path)

		self.opath = path
		self.path = re.compile( "^" + path + "([?].*)?$") 
		self.func = func
		if cache != None:
			self.cache = WebPageCache(cache)
		else:
			self.cache = None
		self.lock = threading.Lock()


	def secure_render(self):

		try:
			return self.func()
		except Exception as e:
			logger.error("Error rendering page '%s' %s " % (self.opath,e))
			import traceback
			print traceback.format_exc()
			return error_page(500,'Error rendering page :anguished: <br> %s' % e)


	def render(self):

		with self.lock:

			if self.cache == None:		
				return self.secure_render()

			cache = self.cache.get()
			if cache == None:
				cache = self.secure_render()
				self.cache.set(cache)

			return cache






#
# This class contains the url and their handler method
# To add a new url, the web object is used as a decorator.
# 
class WebServerDispatcher:

	#
	# Constructor
	#
	def __init__(self):

		# Lists with the urls and ssl urls
		self.urls = []
		self.surls = []


	#
	# Decorator function
	# This method retrieves the decotator parameters
	# and returns the object
	#
	def url(self,path=None,ssl=False,cache=None):

		self.path = path
		self.cache = cache
		self.ssl = ssl

		return self

	#
	# This method is called with the function passed to the decorator
	#
	def __call__(self,func):

		page = WebPage(self.path,func,self.cache)

		if self.ssl == True:
			self.urls.append(page)
		elif self.ssl == False:
			self.surls.append(page)
		else:
			self.urls.append(page)
			self.surls.append(page)

		# Return the function so it can be used outside
		return func


	def dispatch(self,path,ssl):

		if ssl == True:
			urls = self.urls
		else:
			urls = self.surls

		for page in urls:
			if page.path.match(path):
				return page
		return None




web = WebServerDispatcher()




class WebRequestProxy(object):

	PRIVATE = ['_object','_getid','getCookie','setCookie','getParam','send_header','objs']


	def __init__(self):
		self.objs = dict()

	def _getid(self):
		thread = threading.current_thread()
		return thread.ident


	# Get the real object that stores the attributes
	def _object(self):

		# Create the object if not exists
		if not self._getid() in self.__dict__['objs']:
			self.objs[self._getid()] = dict()

		# Return the object
		return self.objs[self._getid()]


	def __getattr__(self,name):

		if name in WebRequestProxy.PRIVATE:
			return object.__getattr__(self, name)
		return self._object()[name]


	def __setattr__(self,name,value):

		if name in WebRequestProxy.PRIVATE:
			object.__setattr__(self, name, value)
			return
		self._object()[name] = value


	def getCookie(self,name):
		
		try:
			c = Cookie.SimpleCookie()
			c.load(self.handler.headers.get('Cookie'))
			return c[name].value
		except:
			return None

	def getParam(self,name):

		# It's post
		if self.post_form != None:
			return self.post_form.getvalue(name, 0)

		# It's get
		query = urlparse.parse_qs(urlparse.urlparse(self.handler.path).query)

		if name in query:
			return query[name][0]
		else:
			return None

	def send_header(self,name,value):
		self.headers.append((name,value))





webrequest = WebRequestProxy()




def render_template(template,context):

	#context['rendertime'] = time.time() - webrequest.start_time

	template = env.get_template(template)
	return template.render(context)




def error_page(response,message):


	webrequest.response = response


	context = Context()
	context['title'] = 'Error %d' % response 
	context['message'] = message 


	return render_template('error.html',context)




#
# The WebServer class is the handler that manages the web requests (GET/POST)
#
class WebServer:


	def __init__(self):

		# Change the server and sys versions
		self.server_version = 'Tablon/' + CONFIG.version
		self.sys_version = ""

	def do_GET(self):
		self.do_GET_or_POST()

	def do_POST(self):
		self.do_GET_or_POST(True)


	#
	# Unified GET and POST
	#
	def do_GET_or_POST(self,post=False):

		webrequest.start_time = time.time()


		ssl = self.server.using_ssl;

		# Get the page object
		page = web.dispatch(self.path,ssl)

		if post:
			form = cgi.FieldStorage(
                fp=self.rfile,
                headers=self.headers,
                environ={'REQUEST_METHOD':'POST',
                         'CONTENT_TYPE':self.headers['Content-Type'],
                         })
			webrequest.post_form = form
		else:
			webrequest.post_form = None


		if page == None:
			self.serve_file()
			return

		try:
			webrequest.original_port = utils.get_originalport(self.rfile._sock)
		except Exception as e:
			print e
			webrequest.original_port = None

		if webrequest.original_port == None:
			webrequest.original_port = CONFIG.port


		# Default atributes
		webrequest.headers = []
		webrequest.handler = self
		webrequest.path = self.path
		webrequest.content_type = 'text/html'
		webrequest.is_binary = False
		webrequest.response = 200
		webrequest.cachecontrol = False

		#
		# Render the page (This can chagne the attributes)
		#
		content = page.render()

	
		self.send_response(webrequest.response)
		self.send_header('Content-type',webrequest.content_type)

		if webrequest.cachecontrol:
			self.write_cache_control()

		for name, value in webrequest.headers:
			self.send_header(name, value)


		self.end_headers()
		if content != None:
			if not webrequest.is_binary:
				self.wfile.write(content.encode("utf-8"))
			else:
				self.wfile.write(content)








	#
	# Send a static file
	#
	def serve_file(self):

		# Get the name (remove the / symbol)
		name = self.path[1:]

		# Disable other folders
		name = name.replace('..','')

		# Check if it really exists
		fname = 'httpdocs/' + name 
		if not os.path.isfile(fname):
			self.send_error_page(404,'Not Found "%s"' % name)
			return

		# Send the file
		self.send_response(200)
		if name.endswith('.css'):
			self.send_header('Content-type','text/css')
		elif name.endswith('.js'):
			self.send_header('Content-type','application/javascript')
		elif name.endswith('.html'):
			self.send_header('Content-type','text/html')
		elif  name.endswith('.txt'):
			self.send_header('Content-type','text/plain')
		elif name.endswith('.ico'):
			self.send_header('Content-type','image/x-icon')
		elif name.endswith('.mf'):
			self.send_header('Content-type','text/cache-manifest')
		elif name.endswith('.svg'):
			self.send_header('Content-type','image/svg+xml')
		else:
			self.send_header('Content-type','application/octet-stream')


		self.write_cache_control()

		# End the headers
		self.end_headers()


		# Send file
		with open(fname,'rb') as f:
			content = f.read()
		self.wfile.write(content)



	def send_error_page(self,number,message):

		self.send_response(number)
		self.send_header('Content-type','text/html')
		self.end_headers()

		template = env.get_template('error.html')
		context = Context()
		context['title'] = 'Error %d' % number 
		context['message'] = message 

		self.wfile.write(template.render(context).encode("utf-8"))



	def write_cache_control(self):

		dt = datetime.now() + timedelta(hours=CONFIG.expire_html)
		htmldt = dt.strftime('%a, %d %b %Y %H:%M:%S GMT')
		self.send_header('Expires',htmldt)













